/* 
 * Network accounting
 * process.c - process packets
 * (C) 1994, 1995 Ulrich Callmeier
 *
 * modified April/98 by Tom Brown - tbrown@baremetal.com 
 *  added hash table to speed up register_packet
 *
 * modified January/2002 by Boril Yonchev /BASHbug/ - by@hibit.net
 *  added mysql support for compact log
 */

#include <sys/time.h>
#include <sys/wait.h>
#include "netacct.h"

#include <stdlib.h>
#include <unistd.h>

#if defined (_LINUX_)
#include <malloc.h>
#endif

#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <utmp.h>
#include <sys/ioctl.h>
#include <sys/stat.h>

#ifdef REMAP_MASQUERADE
#include <linux/ip_masq.h>
#define PROC_MASQ_FILENAME "/proc/net/ip_masquerade"
#endif
/* for mysql connect */
/*#include "mysqldb.h"*/
#include <mysql.h>
#include <syslog.h>

char *rcs_revision_process_c = "$Revision: 1.37 $";

volatile int running;
struct statistics *packets;

#ifndef HASH_SIZE
#define HASH_SIZE 4096
#endif
#define HASH(s,d) ((s &0xfff)^(d &0xfff))


static struct ipdata *plist[HASH_SIZE]; /* data being collected */
static struct ipdata *olist[HASH_SIZE]; /* data being written */
unsigned long int plistsize;
unsigned long int olistsize;
static volatile sig_atomic_t lck;
static volatile sig_atomic_t writing;
static volatile sig_atomic_t dumping;
volatile pid_t writepid;
volatile pid_t dumppid;

volatile int may_write;

int err_delay, max_err_delay;

volatile time_t now;

/* statistics */
unsigned int list_compares, list_lookups;

#ifdef REMAP_MASQUERADE
void remap_masq(struct ipdata *p)
{
    FILE *fd;
    char buf[257];
    char *protocol;
    char searchtext[20];
    int plen;

    if(p->proto == 6)
	protocol = "TCP";
    else if(p->proto == 17) 
	protocol = "UDP";
    else if(p->proto == 1)
	protocol = "ICMP";
    else
	return;

    plen = strlen(protocol);
    sprintf(searchtext, "%08lX:%04X %04X",ntohl(p->dst), 
	p->dstport, p->srcport);
    if((fd=fopen(PROC_MASQ_FILENAME,"r")) == NULL)
	return;

    while(fgets(buf,257,fd)) {
	  if((strncmp(protocol,buf,plen) == 0) &&
	  (strncmp(searchtext, buf+plen+15, 18) == 0)) {
	    p->msqsrc = htonl(strtoul(buf+plen+1, NULL, 16));
	    p->msqport = strtoul(buf+plen+10, NULL, 16);
	    break;
	}
    }
    fclose(fd);
    return;
}

/* BJE:HACK:remap masqueraded entries *BOTH* ways! */
void remap_masq_2(struct ipdata *p)
{
    FILE *fd;
    char buf[257];
    char *protocol;
    char searchtext[20];
    int plen;

    if(p->proto == 6)
	protocol = "TCP";
    else if(p->proto == 17) 
	protocol = "UDP";
    else if(p->proto == 1)
	protocol = "ICMP";
    else
	return;

    plen = strlen(protocol);
    sprintf(searchtext, "%08lX:%04X %04X",ntohl(p->src), 
	p->srcport, p->dstport);
    if((fd=fopen(PROC_MASQ_FILENAME,"r")) == NULL)
	return;

    while(fgets(buf,257,fd)) {
	  if((strncmp(protocol,buf,plen) == 0) &&
	  (strncmp(searchtext, buf+plen+15, 18) == 0)) {
	    p->dst = htonl(strtoul(buf+plen+1, NULL, 16));
	    p->dstport = strtoul(buf+plen+10, NULL, 16);
	    DEBUG(DBG_STATE, sprintf(buf, "MASQ: matched %s", searchtext));
	    break;
	}
    }
    fclose(fd);
    return;
}
#endif

void reopen_socket(void)
{
    int save1, save2;
    
    /* critical section */
    save1 = may_write;
    may_write = 0;
    save2 = lck;
    lck = 1;
    /* end */
    
    exit_capture();
    init_capture();

    lck = save2;
    may_write = save1;
}


void do_acct()
{
  int i;

  packets = malloc(sizeof(struct statistics));

  if(!packets)
    {
      syslog(LOG_ERR,"out of memory");
      daemon_stop(0);
    }
  
  packets->ignored = packets->netignored = packets->ip = packets->local = 0;
  packets->ip_icmp = packets->ip_tcp = packets->ip_udp = packets->ip_other = 0;
  packets->notdev = packets->unenc = 0;

  for (i =0; i < HASH_SIZE; i++) {
     olist[i] = plist[i] = NULL;
     /*  olist = plist = NULL; */
  }
  olistsize = plistsize = 0;
  lck = writing = dumping = 0;

  max_err_delay = cfg -> err_delay;
  err_delay = 0;

  list_lookups = list_compares = 0;

  may_write = 1;

  now = time(NULL);

  alarm(1);
  running=1;

  packet_loop();

}

#define DISABLED(n) (cfg->disabled & BITMASK(n))

void register_packet(unsigned long int src,unsigned long int dst, unsigned char proto, unsigned short srcport, unsigned short dstport, int size, char *devname, char *user)
{
  int hash_val = HASH(src,dst);

  if(lck==0)
    {
      struct ipdata *p;
      DEBUG(DBG_ANNOYING, sprintf(dbg, "lck = 1\n"));
      lck = 1;
      p = plist[ hash_val ];
      list_lookups++;
      while(p)
	{
	  list_compares++;
	  if( (DISABLED(DIS_PROTO) || (p->proto == proto)) 
           && ( DISABLED(DIS_SRC) || (p->src == src)) 
           && (DISABLED(DIS_DST) || (p->dst == dst)) 
           &&  (DISABLED(DIS_SRCPORT) || (p->srcport == srcport)) 
           && (DISABLED(DIS_DSTPORT) || (p->dstport == dstport)) 
           && (strcmp(p->devname, devname)==0))
	    {
	      p->bytes +=size;
	      p->when = now;
	      p->count++;
	      lck = 0;
	      DEBUG(DBG_ANNOYING, sprintf(dbg, "lck = 0\n"));
	      return;
	    }
	  p = p->next;
	}
      p = malloc(sizeof(struct ipdata));
      if(p == NULL)
	{
	  packets -> dropped++;
	  lck = 0;
	  DEBUG(DBG_ERR, sprintf(dbg, "out of memory\n"));
	  DEBUG(DBG_ANNOYING, sprintf(dbg, "lck = 0\n"));
	  return;
	}
      plistsize++;
      p -> src = src;
      p -> dst = dst;
      p -> proto = proto;
      p -> srcport = srcport;
      p -> dstport = dstport;
#ifdef REMAP_MASQUERADE
      p -> msqsrc = 0;
      p -> msqport = 0;
      if((src == cfg->masqif) && 
	 (srcport >= PORT_MASQ_BEGIN) &&
	 (srcport <= PORT_MASQ_END)) {
	    remap_masq(p);
      }
      /* BJE:HACK:MASQ */
      if ((dst == cfg->masqif)) remap_masq_2(p);
#endif
      p -> bytes = size;
      p -> count = 1;
      p -> devname = strdup(devname);
      if (user)
	p -> user = strdup(user);
      else
	p -> user = NULL;
      p -> next = plist[ hash_val ];
      p -> when = now;
      p -> when_start = now;  /* by kad, duration */
      
#ifdef TCP_USER_INFO
      /* NEW !!! get user for tcp only */
      if((proto == IP_TCP) && (p->user == 0) && (get_tcp_info(p, &uid)==1))
	{
	  char tmp[16];
	  sprintf(tmp, "<%d>", uid);
	  p -> user = strdup(tmp);
	}
#endif
      
      plist[ hash_val ] = p;
      lck = 0;
      DEBUG(DBG_ANNOYING, sprintf(dbg, "lck = 0\n"));
    }
  else
    {
      packets->dropped++;
    }
}

int do_write_list(FILE *f, struct ipdata *list[])
{
    struct ipdata *p;
    int i;
    int peer_flag=0; /* peering check */
#ifdef HUMAN_READABLE_TIME
    struct tm *tm1;
    time_t tmm;
#endif 
    
    for( i=0; i < HASH_SIZE; i++) {
        p = list[i];

    while(p)
    {

#ifdef HUMAN_READABLE_TIME
	  time(&tmm);
	  tm1 = localtime(&tmm); 
	  if(fprintf(f, "%02d/%02d/%02d %02d:%02d:%02d",tm1->tm_year,tm1->tm_mon+1,tm1->tm_mday,
		     tm1->tm_hour,tm1->tm_min,tm1->tm_sec)<0) return 1;
#else
	  if(fprintf(f, "%lu",time(0))<0) return 1;
#endif	


   
	    if(!DISABLED(DIS_PROTO)) 
	     if(fprintf(f, "\t%d",p->proto)<0) return 1;

#ifdef REMAP_MASQUERADE
	    if(p->msqport == 0) {
#endif
	      if(!DISABLED(DIS_SRC)) 
	       if(fprintf(f, "\t%s",intoa(p->src))<0) return 1;
/*	      syslog(LOG_DEBUG,"p->src = %s",intoa(p->src));*/

	      if(!DISABLED(DIS_SRCPORT)) 
	       if(fprintf(f, "\t%d",p->srcport)<0) return 1;
#ifdef REMAP_MASQUERADE
	    }
	    else {
	      if(!DISABLED(DIS_SRC)) if(fprintf(f, "\t%s",intoa(p->msqsrc))<0) return 1;

	      if(!DISABLED(DIS_SRCPORT)) if(fprintf(f, "\t%d",p->msqport)<0) return 1;
	    }		
#endif

	    if(!DISABLED(DIS_DST)) 
	     if(fprintf(f, "\t%s",intoa(p->dst))<0) return 1;
/*	     syslog(LOG_DEBUG,"p->dst = %s",intoa(p->dst));*/

	    if(!DISABLED(DIS_DSTPORT)) 
	     if(fprintf(f, "\t%d",p->dstport)<0) return 1;

	    if(!DISABLED(DIS_COUNT)) 
	     if(fprintf(f, "\t%u",p->count)<0) return 1;

	    if(!DISABLED(DIS_BYTES)) 
	     if(fprintf(f, "\t%lu",p->bytes)<0) return 1;

	    if(!DISABLED(DIS_DEV)) 
	     if(fprintf(f, "\t%s",p->devname)<0) return 1;

	    if(!DISABLED(DIS_USER)) 
	     if(fprintf(f, "\t%s",p->user?p->user:"unknown")<0) return 1;
	   
	    /* add duration field by geroy */
	    if(!DISABLED(DIS_DUR))
	      if(fprintf(f, "\t%i",p->when-p->when_start)<0) return 1;

	    /* peering flag, check if it is in peering net 
	     * by geroy */
            if(onnetlist(p->src,pcfg->peering_addr) || onnetlist(p->dst,pcfg->peering_addr))
	      peer_flag = 1;
	    else
	      peer_flag = 0;
		    
	    if(!DISABLED(DIS_FLAG))
	      if(fprintf(f, "\t%i", peer_flag)<0) return 1;

	    if(fprintf(f, "\n")<0) return 1;
   
      p = p->next;
   }

}
    if(fclose(f)<0)
	{
	    return 1;
	}
	    
    return 0;
}

int do_write_list_mysql(struct ipdata *list[])
{
    struct ipdata *p;
    int i;
    /* --- patch by geroy --- connect to mysql server */
    /* needed for mysql query
     * here is a patch to activate disable options in config file */
    char query[8192];
    char construct[64];
    MYSQL mysql;
    /* patched by BASHbug */
    MYSQL_RES *result;
    MYSQL_ROW r;
    my_ulonglong row;
    unsigned long int old_input, old_output, old_peer_input, old_peer_output;
    unsigned long int old_direct_input, old_direct_output, old_local_input, old_local_output;
    int rc, src_flag=0, peer_flag=0;
    char spyip[18];
    char TIME_MASK[] = "DATE_FORMAT(NOW(),'%Y-%m-%d %H:00:00')";
    /* patched by BASHbug */
    char spek_src[18];
    char spek_dst[18];
    
    
    /* connect to mySQL db with user/pass. Note that you must put it
     * mysqldb.h your user/pass/mysql_host/database */

    mysql_init(&mysql);
/*    if(!mysql_real_connect(&mysql,MYSQL_HOST,MYSQL_USER,MYSQL_PASS,MYSQL_BASE,0,NULL,0))*/
    /* patched by d.sadilek@globalview.de, use mysql_port from configuration */
    if(!mysql_real_connect(&mysql,cfg->mysql_host,cfg->mysql_user,cfg->mysql_password,cfg->mysql_database,cfg->mysql_port,NULL,0))
    {
	     syslog(LOG_DEBUG,"mySQL Error: %s",mysql_error(&mysql));
	     return 1;
    }

    for( i=0; i < HASH_SIZE; i++) {
        p = list[i];

    while(p)
    {
     /* I really don't why intoa() cause problem but this should solve the 
      * problem (may be my PC is too slow but when I remove this two lines
      * and put intoa(p->src) and intoa(p->dst) instead of spek_src and 
      * spek_dst in mysql query, the result is that src = dst which is 
      * not true. If you can, fix it :) */
	    
     strncpy(spek_src,intoa(p->src),16);
     strncpy(spek_dst,intoa(p->dst),16);

     /* patched by Renegade */
    
     /* patched by BASHbug */
     
     DEBUG(DBG_STATE,sprintf(dbg,"analiz for src %s, dst %s\n",spek_src,spek_dst));

     if(onnetlist(p->src,cfg->compactnet) || onnetlist(p->dst,cfg->compactnet))
     {
	     /* check if it is from compactnet ip group */

	    if(onnetlist(p->src,cfg->compactnet)) 
	      {

		   if ( onnetlist(p->src, cfg->ournet) )
          		{

		            if (onnetlist(p->dst, cfg->ournet))
                		{
		                    peer_flag = 3; /* Local Traffic */
                		}
		            else if (onnetlist(p->dst, cfg->direct_peer))
                		{
		                    peer_flag = 2; /* Direct peer */
                		}
		            else if (onnetlist(p->dst, pcfg->peering_addr) && !onnetlist(p->dst, cfg->ournet))
                		{
		                    peer_flag = 1; /* Peering traffic */
                		}
		            else
                		    peer_flag = 0; /* Other traffic */
          		}
		      
	          strncpy (spyip,spek_src,16);
		  src_flag=1;
	      }
	    else 
	      {

		    if ( onnetlist(p->dst, cfg->ournet) )
          		{
		            if (onnetlist(p->src, cfg->ournet))
		                {
                		    peer_flag = 3; /* Local traffic */
		                }
		            else if (onnetlist(p->src, cfg->direct_peer))
                		{
		                    peer_flag = 2; /* Direct peer */
                		}
		            else if (onnetlist(p->src, pcfg->peering_addr) && !onnetlist(p->src, cfg->ournet))
                		{
		                    peer_flag = 1; /* Peering traffic */
                		}
		            else
                		    peer_flag = 0;
         		}

	          strncpy (spyip,spek_dst,16);
		  src_flag=0;
	      }
	    /* syslog(LOG_INFO,"flag = %i", flag); */

	    sprintf(query, "SELECT input,output,peer_input,peer_output,direct_input,direct_output,local_input,local_output FROM %s WHERE ip='%s' AND time=%s","traffic", spyip,TIME_MASK);

	    
	     /* syslog(LOG_INFO,"query = %s",query); */
       
		rc = mysql_query(&mysql, query);
       		result = mysql_store_result(&mysql);
       		row = mysql_num_rows(result);
		
	    /* syslog(LOG_INFO,"row = %li",row); */
		
       	    if (row) /* if it is old compact info */
            {
    	    	r = mysql_fetch_row(result); /* get query result */
		     {
                        old_input = atol(r[0]);
                        old_output = atol(r[1]);
                        old_peer_input = atol(r[2]);
                        old_peer_output = atol(r[3]);
                        old_direct_input = atol(r[4]);
                        old_direct_output = atol(r[5]);
                        old_local_input = atol(r[6]);
                        old_local_output = atol(r[7]);
		     }

                     old_input+=(!src_flag && peer_flag==0)?p->bytes:0;
                     old_output+=(src_flag && peer_flag==0)?p->bytes:0;
                     old_peer_input+=(!src_flag && peer_flag==1)?p->bytes:0;
                     old_peer_output+=(src_flag && peer_flag==1)?p->bytes:0;
                     old_direct_input+=(!src_flag && peer_flag==2)?p->bytes:0;
                     old_direct_output+=(src_flag && peer_flag==2)?p->bytes:0;
                     old_local_input+=(!src_flag && peer_flag==3)?p->bytes:0;
                     old_local_output+=(src_flag && peer_flag==3)?p->bytes:0;


  		/* syslog(LOG_INFO,"last record is - %li",old_trafic); */
		/* syslog(LOG_INFO,"sum = %li",old_trafic); */

            	DEBUG(DBG_STATE,sprintf(dbg,"upate info for ip %s, peer_flag=%u\n",spyip,peer_flag));

                sprintf(query,"UPDATE %s SET input='%li', output='%li', peer_input='%li', peer_output='%li', direct_input='%li', direct_output='%li', local_input='%li',local_output='%li' WHERE ip='%s' AND time=%s", "traffic", old_input,old_output, old_peer_input, old_peer_output,old_direct_input, old_direct_output,old_local_input,old_local_output, spyip,TIME_MASK);

	  
	    /* 	syslog(LOG_INFO,"query = %s",query);*/
	   
	    }	    
    	    else  /* if it new record in compact info */
            {

                    old_input=(!src_flag && peer_flag==0)?p->bytes:0;
                    old_output=(src_flag && peer_flag==0)?p->bytes:0;
                    old_peer_input=(!src_flag && peer_flag==1)?p->bytes:0;
                    old_peer_output=(src_flag && peer_flag==1)?p->bytes:0;
                    old_direct_input=(!src_flag && peer_flag==2)?p->bytes:0;
                    old_direct_output=(src_flag && peer_flag==2)?p->bytes:0;
                    old_local_input=(!src_flag && peer_flag==3)?p->bytes:0;
                    old_local_output=(src_flag && peer_flag==3)?p->bytes:0;

		/* syslog(LOG_INFO,"insert info for ip %s",spyip); */
            	DEBUG(DBG_STATE,sprintf(dbg,"upate info for ip %s\n",spyip));

	sprintf(query,"INSERT INTO %s (ip,time, input,output,peer_input,peer_output,direct_input,direct_output,local_input,local_output) VALUES ( '%s',%s,'%li','%li','%li','%li','%li','%li','%li','%li')", "traffic", spyip, TIME_MASK, old_input,old_output,old_peer_input,old_peer_output,old_direct_input,old_direct_output, old_local_input,old_local_output);


   		 /*  syslog(LOG_INFO,"query = %s",query); */		

	    }
     }
     else
     {
	 /* Now we will construct our query using disable options
	  * from config file ... */

	     /* beginning of query */
       sprintf(query, "INSERT INTO %s (time, data, duration, protocol, src_ip, src_port, dst_ip, dst_port, packet, data_size, device, user, peer) VALUES ( DATE_FORMAT(FROM_UNIXTIME(%li),'%%T'), FROM_UNIXTIME(%li)", cfg->mysql_table, p->when, p->when);
       
       if(!DISABLED(DIS_DUR))
          sprintf(construct, ", %li", p->when - p->when_start);
       else
	  sprintf(construct, ", 0");

       /* add 'duration' to query */
       strcat(query, construct);

       if(!DISABLED(DIS_PROTO))
          sprintf(construct, ",%i", p->proto);
       else
	  sprintf(construct, ", 0");

       /* add 'protocol' to query */
       strcat(query, construct);

       if(!DISABLED(DIS_SRC))
	  sprintf(construct, ",\'%s\'", intoa(p->src));
       else
	  sprintf(construct, ", \'\'");

       /* add 'src ip address' to query */
       strcat(query, construct);

       if(!DISABLED(DIS_SRCPORT))
	  sprintf(construct, ",%i", p->srcport);
       else
	  sprintf(construct, ", 0");

       /* add 'src port' to query */
       strcat(query, construct);

       if(!DISABLED(DIS_DST))
	  sprintf(construct, ",\'%s\'", intoa(p->dst));
       else
	  sprintf(construct, ", \'\'");

       /* add 'dst ip address' to query */
       strcat(query, construct);
       
       if(!DISABLED(DIS_DSTPORT))
	  sprintf(construct, ",%i", p->dstport);
       else
	  sprintf(construct, ", 0");

       /* add 'dst port' to query */
       strcat(query, construct);

       if(!DISABLED(DIS_COUNT))
	  sprintf(construct, ",%li", p->count);
       else
	  sprintf(construct, ", 0");

       /* add 'packet' count to query */
       strcat(query, construct);

       if(!DISABLED(DIS_BYTES))
	  sprintf(construct, ",%li", p->bytes);
       else
	  sprintf(construct, ", 0");

       /* add 'bytes' to query */
       strcat(query, construct);

       if(!DISABLED(DIS_DEV))
	  sprintf(construct, ", \'%s\'", p->devname);
       else
	  sprintf(construct, ", \'\'");

       /* add 'device name' to query */
       strcat(query, construct);

       if(!DISABLED(DIS_USER))
	  sprintf(construct, ", \'%s\'", p->user?p->user:"unknown");
       else
	  sprintf(construct, ", \'\'");

       /* add 'username' to query */
       strcat(query, construct);

       if(!DISABLED(DIS_FLAG))
	  sprintf(construct, ", %i", peer_flag);
       else
	  sprintf(construct, ", 0");

       /* add 'peer_flag' to query
	* almost useless .. but :) */
       strcat(query, construct);

	    
       sprintf(construct, " )\0");

       /* and add ')' to default query */
       strcat(query, construct);

	       
	 /*  syslog(LOG_INFO,"insert info for ip %s",spyip); */
            DEBUG(DBG_STATE,sprintf(dbg,"insert full info for ip %s\n",spyip));

     }
         
     DEBUG(DBG_STATE,sprintf(dbg,"%s\n",query));
 /*    syslog(LOG_INFO,"query is %s",query); */
     
 /*  END patch by BASHbug */

     /* Here goes mysql query, to insert data into mySQL db */
     /* and actual writing to mySQL db*/
     
     mysql_query(&mysql, query);

     p = p->next;
    }     
  }
    /* close connection to mySQL */
    mysql_close(&mysql);

/* patched by BASHbug
     if(fclose(f)<0)
         {
               return 1;
         }
*/      		 
	    
    return 0;
}


static int pfd1[2], pfd2[2];

void TELL_WAIT_INIT(void)
{
  if(pipe(pfd1) < 0 || pipe(pfd2) < 0)
    {
      syslog(LOG_ERR, "pipe error: %s", strerror(errno));
      DEBUG(DBG_ERR, sprintf(dbg,"pipe error: %s\n", strerror(errno)));
    }
}

void TELL_WAIT_EXIT(void)
{
  if(close(pfd1[0])!=0)
    {
      syslog(LOG_ERR, "pipe close error: %s", strerror(errno));
      DEBUG(DBG_ERR, sprintf(dbg,"pipe close error: %s\n", strerror(errno)));
    }

  if(close(pfd1[1])!=0)
    {
      syslog(LOG_ERR, "pipe close error: %s", strerror(errno));
      DEBUG(DBG_ERR, sprintf(dbg,"pipe close error: %s\n", strerror(errno)));
    }

  if(close(pfd2[0])!=0)
    {
      syslog(LOG_ERR, "pipe close error: %s", strerror(errno));
      DEBUG(DBG_ERR, sprintf(dbg,"pipe close error: %s\n", strerror(errno)));
    }

  if(close(pfd2[1])!=0)
    {
      syslog(LOG_ERR, "pipe close error: %s", strerror(errno));
      DEBUG(DBG_ERR, sprintf(dbg,"pipe close error: %s\n", strerror(errno)));
    }
}

void TELL_PARENT(void)
{
  if(write(pfd2[1], "c", 1) != 1)
    {
      syslog(LOG_ERR, "write error: %s", strerror(errno));
      DEBUG(DBG_ERR, sprintf(dbg,"write error: %s\n", strerror(errno)));
    }
}

void WAIT_PARENT(void)
{
  char c; int n;

 again:

  if((n = read(pfd1[0], &c, 1)) != 1)
    {
      if((n == -1) && (errno == EINTR)) goto again;
      syslog(LOG_ERR, "read error: %s", strerror(errno));
      DEBUG(DBG_ERR, sprintf(dbg,"read error: %s\n", strerror(errno)));
    }
  if(c!='p')
    {
      syslog(LOG_ERR, "WAIT_PARENT: incorrect data");
      DEBUG(DBG_ERR, sprintf(dbg,"WAIT_PARENT: incorrect data\n"));
    }
}  

void TELL_CHILD(void)
{
  if(write(pfd1[1], "p", 1) != 1)
    {
      syslog(LOG_ERR, "write error: %s", strerror(errno));
      DEBUG(DBG_ERR, sprintf(dbg,"write error: %s\n", strerror(errno)));
    }
}

void WAIT_CHILD(void)
{
  char c; int n;

 again:

  if((n = read(pfd2[0], &c, 1)) != 1)
    {
      if((n == -1) && (errno == EINTR)) goto again;
      syslog(LOG_ERR, "read error: %s", strerror(errno));
      DEBUG(DBG_ERR, sprintf(dbg,"read error: %s\n", strerror(errno)));
    }
  if(c!='c')
    {
      syslog(LOG_ERR, "WAIT_CHILD: incorrect data");
      DEBUG(DBG_ERR, sprintf(dbg,"WAIT_CHILD: incorrect data\n"));
    }
}  

/* write and clear olist */
void write_list(void)
{
/* patched by BASHbug */
    FILE *f; /* needed only when working with file */
    
    char tmpn[255];
    int i;

    while( (writepid = fork()) < 0) sleep(1);
    if (writepid!=0) return;

    /* Here goes the child */

    TELL_PARENT();
    WAIT_PARENT();

    DEBUG(DBG_STATE, sprintf(dbg, "write child: synchronized with parent\n"));

    sprintf(tmpn, "/tmp/nacctd.write.%d", (int) getpid());
    creat(tmpn, S_IRUSR);

    openlog("nacctd (write)", 0, LOG_DAEMON);

    DEBUG(DBG_STATE, sprintf(dbg, "* write process %d forked\n", (int) getpid()));
/* patched by BASHbug
    f = fopen(cfg->filename, "a");
    if(f==NULL)
	{
	    unlink(tmpn);
	    syslog(LOG_ERR, "error opening file %s: %m\n",cfg->filename);
	    exit(1);
	}
*/
    /* begin of check for mysql or file */ 
    /* if databes = file (from config file then work with
     * /var/log/net-acct instead of mysql */
    
   if(strcasecmp(cfg->database, "file")==0)
   {
   /* open the file for appending */
    f = fopen(cfg->filename, "a");
    if(f==NULL)
    {
	    unlink(tmpn);
	    syslog(LOG_ERR, "error opening file %s: %m\n",cfg->filename);
	    exit(1);
    }
    if(do_write_list(f, olist) != 0)
    {
	    unlink(tmpn);
	    syslog(LOG_ERR, "error writing to file %s: %m\n", cfg->filename);
	    exit(1);
    }
   }
   /* else if database = mysql - work with mysql*/
   else
   {
	    
    if(do_write_list_mysql(olist) != 0)
	{
	    unlink(tmpn);
	    syslog(LOG_ERR, "error writing to file %s: %m\n", cfg->filename);
	    exit(1);
	}
    } /* end of check for mysql or file*/
   
    /* memory leak? TAB */
    for (i=0; i < HASH_SIZE; i++) 
       olist[i] = NULL;

    unlink(tmpn);

    DEBUG(DBG_STATE, sprintf(dbg, "* write finished, count = %ld\n", olistsize));

    exit(0);
}

void dump_curr_list(void)
{
    FILE *f;
    char tmpn[255];
    int i;

    while( (dumppid = fork()) < 0) sleep(1);
    if (dumppid!=0) return;

    TELL_PARENT();
    WAIT_PARENT();

    DEBUG(DBG_STATE, sprintf(dbg, "dump child: synchronized with parent\n"));

    /* Here goes the child */

    sprintf(tmpn, "/tmp/nacctd.dump.%d", (int) getpid());
    creat(tmpn, S_IRUSR);

    openlog("nacctd (dump)", 0, LOG_DAEMON);

    DEBUG(DBG_STATE, sprintf(dbg, "* dump process %d forked\n", (int) getpid()));

    if(plistsize == 0)
	{
	    unlink(tmpn);
	    unlink(cfg->dumpname);
	    DEBUG(DBG_STATE, sprintf(dbg, "* dump finished, dump empty\n"));
	    exit(0);
	}
/* Patched by BASHbug
    f = fopen(cfg->dumpname, "w");
    hmm?! must be "w"
*/    
    f = fopen(cfg->dumpname, "w");
    if(f==NULL)
	{
	    unlink(tmpn);
	    syslog(LOG_ERR, "error opening file %s: %m\n",cfg->dumpname);
	    exit(1);
	}

    if(do_write_list(f, plist) != 0)
	{
	    unlink(tmpn);
	    syslog(LOG_ERR, "error writing to file %s: %m\n", cfg->dumpname);
	    exit(1);
	}

    for (i=0; i < HASH_SIZE; i++) 
       plist[i] = NULL;

    unlink(tmpn);

    DEBUG(DBG_STATE, sprintf(dbg, "* dump finished, count = %ld\n", plistsize));

    exit(0);
}


void child_finished(int sig)
{
    int status;
    pid_t pid;

    DEBUG((DBG_SIGNAL | DBG_STATE), sprintf(dbg, "-> got signal %d, handling\n", sig));

    while((pid = waitpid((pid_t) -1, &status, WNOHANG)) != 0)
	{
	  DEBUG(DBG_SIGNAL, sprintf(dbg, "  waitpid returned %d, status = %d, errno = %d\n", pid, status, errno));
	     
	  if(pid == -1)
	    {
	      if(errno == ECHILD)
		break; /* no child processes */
	      DEBUG(DBG_SIGNAL, sprintf(dbg, "waitpid: signaled error: %s\n", strerror(errno)));
	    }
	      
	  if((pid == writepid) || (pid == dumppid))
	    {
	      if(WIFEXITED(status))
		{
		  if(WEXITSTATUS(status)==0)
		    {
		      if(pid == writepid)
			{
			  writing = 0;
			  DEBUG((DBG_SIGNAL | DBG_STATE), sprintf(dbg, "  set writing to 0\n"));
			}
		      else
			{
			  dumping = 0;
			  DEBUG((DBG_SIGNAL | DBG_STATE), sprintf(dbg, "  set dumping to 0\n"));
			}
		    }
		  else
		    {
		      syslog(LOG_ERR, 
			     "child %d exited with error status %d.\n",
			     pid, WEXITSTATUS(status));
		      if(pid == writepid)
			{
			  err_delay = max_err_delay;
			  writing = 0;
			  DEBUG((DBG_SIGNAL | DBG_STATE), sprintf(dbg, "  set writing to 0, setting err_delay\n"));
			}
		      else
			{
			  dumping = 0;
			  DEBUG((DBG_SIGNAL | DBG_STATE), sprintf(dbg, "  set dumping to 0, ignored error condition\n"));
			}
		      
		    }
		}
	      else
		{
		  syslog(LOG_ERR,
			 "Huh? Child %d terminated or stopped by signal (%m)\n",
			 pid);
		  if(pid == writepid)
		    {
		      writing = 0;
		      DEBUG((DBG_SIGNAL | DBG_STATE), sprintf(dbg, "  set writing to 0, ignored return code\n"));
		    }
		  else
		    {
		      dumping = 0;
		      DEBUG((DBG_SIGNAL | DBG_STATE), sprintf(dbg, "  set dumping to 0, ignored return code\n"));
		    }
		}
	    }
	  else
	    {
	      syslog(LOG_ERR, "Huh? Child (%d) returned, but not the one we expected (%d, %d)!\n", (int) pid, writepid, dumppid);
	      DEBUG(DBG_STATE, sprintf(dbg, "  unexpected child %d signaled return (writepid = %d, dumppid = %d\n",(int) pid, writepid, dumppid));

	    }
	  DEBUG(DBG_STATE, sprintf(dbg, "  child %d signaled return\n",(int) pid));
	}
    DEBUG((DBG_SIGNAL | DBG_STATE), sprintf(dbg, "<- got signal %d, done handling\n", sig));
}

void alarm_handler(int sig)
{
    static time_t last_check = 0;
    static time_t next_write_log = 0;

    DEBUG( ((sig == SIGALRM) ? DBG_ANNOYING : (DBG_SIGNAL | DBG_STATE)), sprintf(dbg, "got signal %d, handling\n", sig));

    now++;

    if((now - last_check) > 60)
	{
	    time_t nnow;
	    
	    nnow = time(NULL);
	    if(nnow!=now)
		{
		    if((abs(nnow - now) > 2))
			{
			    DEBUG(DBG_MISC, sprintf(dbg, "internal clock corrected (off by %d seconds)\n",(int) (nnow-now)));
			}
		    now = nnow;
		}
	    last_check = now;
	}

    if(now >= next_write_log)
	{
	    write_log(0);
	    next_write_log = now + cfg -> flush;
	}

    alarm(1);
}

void write_log(int force)
{
  struct ipdata *p, *q;
  static struct ipdata *tlist; /* temp */
  int i;

  DEBUG(DBG_STATE, sprintf(dbg, "write_log called\n"));

  if(err_delay!=0)
    {
      err_delay--;
      syslog(LOG_INFO,"flushing delayed due to error\n");
      DEBUG(DBG_STATE, sprintf(dbg, "flushing delayed due to error\n"));
    }
  else if((writing == 0) && (lck == 0) && (may_write == 1)) /* delay if another write cycle is still in progress */
    {
      DEBUG(DBG_STATISTICS, sprintf(dbg, "ignored: %ld netignored: %ld local:%ld ip:%ld unenc:%ld notdev:%ld dropped:%ld\n", 
				    packets->ignored, packets->netignored, packets->local, packets->ip, 
				    packets->unenc, packets->notdev, packets->dropped));
      DEBUG(DBG_STATISTICS, sprintf(dbg, "udp: %ld tcp:%ld icmp:%ld other:%ld\n", 
				    packets->ip_udp, packets->ip_tcp, packets->ip_icmp, 
				    packets->ip_other));
      if(list_lookups != 0)
	{
	  DEBUG(DBG_STATISTICS, sprintf(dbg, "lookups:%d compares:%d compares/lookup:%f\n",
					list_lookups, list_compares, 
					((float) list_compares / (float) list_lookups)));
	}
      DEBUG(DBG_STATE, sprintf(dbg, "lck = 1\n"));
      lck = 1; /* can't update the list now */
      
      DEBUG(DBG_MISC, sprintf(dbg, "Total of %ld entries\n", plistsize));
      
      /* We build two lists:
	 1) olist, which will be written out
	 2) tlist, which will be the new plist (just for this hash row)
	 */
      
      olistsize = 0;
      plistsize = 0;

      for (i = 0; i < HASH_SIZE; i++) {
	 p = plist[i];
	 tlist = NULL;
	 olist[i] = NULL;
	 
	 while(p)
	   {
	     q = p->next;
	     if(((now - p->when) > cfg->fdelay) || force)
	       {
		 p->next = olist[i];
		 olist[i] = p;
		 olistsize++;
	       }
	     else
	       {
		 p->next = tlist;
		 tlist = p;
		 plistsize++;
	       }
	     p = q;
	   }
	 plist[i] = tlist;
      }
      
      if(dumping == 0)
	{
	  dumping = 1;

	  TELL_WAIT_INIT();

	  dump_curr_list();

	  TELL_CHILD();
	  WAIT_CHILD();
	  
	  DEBUG(DBG_STATE, sprintf(dbg, "parent: synchronized with dump child\n"));
	  
	  TELL_WAIT_EXIT();

	  DEBUG(DBG_STATE, sprintf(dbg, "dumppid is %d\n", (int) dumppid));
	}
      
      writing = 1; /* no further writing 'til this is finished */
      
      lck = 0;
      DEBUG(DBG_STATE, sprintf(dbg, "lck = 0\n"));
      
      DEBUG(DBG_MISC, sprintf(dbg, "Split into %ld [hold] and %ld [write] = %ld [total] entries\n", plistsize, olistsize, plistsize + olistsize));
      
      TELL_WAIT_INIT();

      write_list(); /* this forks off a child to do the actual writing */
      
      TELL_CHILD();
      WAIT_CHILD();

      DEBUG(DBG_STATE, sprintf(dbg, "parent: synchronized with write child\n"));
      
      TELL_WAIT_EXIT();

      DEBUG(DBG_STATE, sprintf(dbg, "writepid is %d\n", (int) writepid));
      
      for (i=0; i < HASH_SIZE; i++) {
	 p = olist[i];
	 while(p)
	   {
	     olist[i] = p->next;
	     free(p->devname);
	     if (p->user)
		 free(p->user);
	     free(p);
	     p=olist[i];
	   }
      }
        
      
      DEBUG(DBG_STATE, sprintf(dbg, "done freeing\n"));
    }
  else
    {
      DEBUG(DBG_STATE, sprintf(dbg, "flushing delayed (writing == %d, lck == %d, may_write == %d)\n",writing,lck,may_write));
    }
}

void signal_debug(int sig)
{
  DEBUG(DBG_SIGNAL, sprintf(dbg, "got signal %d, handling\n", sig));
    
  if(sig==SIGUSR1)
    {
      debug_level++;
    }
  else if(sig==SIGUSR2)
    {
      syslog(LOG_DEBUG, "turning off debugging\n");
      debug_level = 0;
    }
  else if(sig==SIGWINCH)
    {
      syslog(LOG_DEBUG,"nacctd, revisions:\n%s\n%s\n%s\n%s\n", 
	     rcs_revision_main_c, rcs_revision_process_c,
	     rcs_revision_config_c, rcs_revision_daemon_c);
    }
  else if(sig==SIGTSTP)
    {
      DEBUG(DBG_STATE, sprintf(dbg, "received SIGTSTP\n"));
      may_write = 0;
    }
  else if(sig==SIGCONT)
    {
      DEBUG(DBG_STATE, sprintf(dbg, "received SIGCONT\n"));
      may_write = 1;
    }
  else if(sig==SIGIOT)
    {
      DEBUG(DBG_STATE, sprintf(dbg, "reopening socket\n"));
      reopen_socket();
    }
  else if(sig == SIGHUP)
  {
      struct ipnetwork *ip,*ip_next;
      struct promisc_device *pr,*pr_next;
      struct headerdat *h,*h_next;
      struct mon_host_struct *mh,*mh_next;
      DEBUG(DBG_STATE, sprintf(dbg, "received SIGHUP, rereading config\n"));
      exit_capture();
      if(cfg) {
/* patched by BASHbug
      	      free(cfg->filename);
*/	      
	free(cfg->dumpname);
	free(cfg->debugname);
	for(ip=cfg->ignorenet;ip;ip=ip_next) {
	    ip_next = ip->next;
	    free(ip);
	}
	for(ip=cfg->dontignore;ip;ip=ip_next) {
	    ip_next = ip->next;
	    free(ip);
	}
	for(ip=cfg->excludenamelookup;ip;ip=ip_next) {
	    ip_next = ip->next;
	    free(ip);
	}
	for(pr=cfg->promisc;pr;pr=pr_next) {
	    pr_next = pr->next;
	    free(pr);
	}
	for(pr=cfg->notdev;pr;pr=pr_next) {
	    pr_next = pr->next;
	    free(pr);
	}
	for(pr=cfg->iflist;pr;pr=pr_next) {
	    pr_next = pr->next;
	    free(pr);
	}
	for(h=cfg->headers;h;h=h_next) {
	    h_next = h->next;
	    free(h);
	}
	for(mh=cfg->hostlist;mh;mh=mh_next) {
	    mh_next = mh->next;
	    free(mh);
	}
	free(cfg);
      }
      cfg = read_config(fname);
      init_capture();
  }
  else
    {
      DEBUG(DBG_SIGNAL, sprintf(dbg, "signal_debug received signal %d, this can't happen\n", sig));
      syslog(LOG_INFO,"signal_debug received signal %d, this can't happen\n", sig);
    }
}

void signal_ignore(int sig)
{
  DEBUG(DBG_SIGNAL, sprintf(dbg, "got signal %d, ignoring\n", sig));
}



